import catboost
from .BaseClassifier import BaseClassifier


class CatBoostClassifier(BaseClassifier):

    def __init__(self):
        super().__init__(
            catboost.CatBoostClassifier,
            {'verbose': 0, 'n_estimators': 1000, 'allow_writing_files': False},
        )



import os
import re
import json
import copy
import math
import shutil
import pickle
import joblib
import warnings
import textwrap
import numpy as np
import pandas as pd
import multiprocessing as mp
from scipy.stats import uniform
from scipy.stats import randint
import matplotlib.pyplot as plt

import sklearn
from sklearn import metrics
from sklearn.model_selection import KFold
from sklearn.model_selection import StratifiedKFold
from sklearn.preprocessing import StandardScaler

from . import Utils
from .AutoML.Sequence import Sequence
from .AutoML.Modelling import Modelling
from .AutoML.DataExploring import DataExploring
from .AutoML.DataProcessing import DataProcessing
from .AutoML.FeatureProcessing import FeatureProcessing
from .GridSearch.BaseGridSearch import BaseGridSearch
from .GridSearch.HalvingGridSearch import HalvingGridSearch
from .GridSearch.OptunaGridSearch import OptunaGridSearch


class Pipeline:

    def __init__(self,
                 target,
                 project='',
                 version=None,
                 mode='regression',
                 objective=None,
                 fast_run=False,

                 # Data Processing
                 num_cols=None,
                 date_cols=None,
                 cat_cols=None,
                 missing_values='interpolate',
                 outlier_removal='clip',
                 z_score_threshold=4,
                 include_output=False,

                 # Feature Processing
                 max_lags=10,
                 max_diff=2,
                 information_threshold=0.99,
                 extract_features=True,

                 # Sequencing
                 sequence=False,
                 back=0,
                 forward=0,
                 shift=0,
                 diff='none',

                 # Initial Modelling
                 normalize=True,
                 shuffle=False,
                 cv_splits=3,
                 store_models=True,

                 # Grid Search
                 grid_search_type='optuna',
                 grid_search_iterations=3,

                 # Production
                 custom_code='',

                 # Flags
                 plot_eda=None,
                 process_data=None,
                 validate_result=None,
                 verbose=1):
        # Starting
        print('\n\n*** Starting Amplo AutoML - %s ***\n\n' % project)

        # Parsing input
        if len(project) == 0:
            self.mainDir = 'AutoML/'
        else:
            self.mainDir = project if project[-1] == '/' else project + '/'
        self.target = re.sub('[^a-z0-9]', '_', target.lower())
        self.verbose = verbose
        self.customCode = custom_code
        self.fastRun = fast_run

        # Checks
        assert mode == 'regression' or mode == 'classification', 'Supported modes: regression, classification.'
        assert isinstance(target, str), 'Target needs to be of type string, key of target'
        assert isinstance(project, str), 'Project is a name, needs to be of type string'
        assert isinstance(num_cols, list), 'Num cols must be a list of strings'
        assert isinstance(date_cols, list), 'Date cols must be a list of strings'
        assert isinstance(cat_cols, list), 'Cat Cols must be a list of strings'
        assert isinstance(shift, int), 'Shift needs to be an integer'
        assert isinstance(max_diff, int), 'max_diff needs to be an integer'
        assert max_lags < 50, 'Max_lags too big. Max 50.'
        assert 0 < information_threshold < 1, 'Information threshold needs to be within [0, 1'
        assert max_diff < 5, 'Max difftoo big. Max 5.'

        # Objective
        if objective is not None:
            self.objective = objective
        else:
            if mode == 'regression:':
                self.objective = 'neg_mean_squared_error'
            elif mode == 'classification':
                self.objective = 'accuracy'
        assert isinstance(objective, str), 'Objective needs to be a string.'
        assert objective in metrics.SCORERS.keys(), 'Metric not supported, look at sklearn.metrics.SCORERS.keys()'
        self.scorer = metrics.SCORERS[self.objective]

        # Params needed
        self.mode = mode
        self.version = version
        self.numCols = num_cols
        self.dateCols = date_cols
        self.catCols = cat_cols
        self.missingValues = missing_values
        self.outlierRemoval = outlier_removal
        self.zScoreThreshold = z_score_threshold
        self.includeOutput = include_output
        self.sequence = sequence
        self.sequenceBack = back
        self.sequenceForward = forward
        self.sequenceShift = shift
        self.sequenceDiff = diff
        self.normalize = normalize
        self.shuffle = shuffle
        self.cvSplits = cv_splits
        self.gridSearchType = grid_search_type
        self.gridSearchIterations = grid_search_iterations
        self.plotEDA = plot_eda
        self.processData = process_data
        self.validateResults = validate_result

        # Instance initiating
        self.X = None
        self.Y = None
        self.colKeep = None
        self.results = None

        # Flags
        self._set_flags()

        # Create Directories
        self._create_dirs()

        # Load Version
        self._load_version()

        # Sub- Classes
        self.DataProcessing = DataProcessing(target=self.target, num_cols=num_cols, date_cols=date_cols,
                                             cat_cols=cat_cols, missing_values=missing_values, mode=mode,
                                             outlier_removal=outlier_removal, z_score_threshold=z_score_threshold,
                                             folder=self.mainDir + 'Data/', version=self.version)
        self.FeatureProcessing = FeatureProcessing(max_lags=max_lags, max_diff=max_diff,
                                                   extract_features=extract_features, mode=mode,
                                                   information_threshold=information_threshold,
                                                   folder=self.mainDir + 'Features/', version=self.version)
        self.Sequence = Sequence(back=back, forward=forward, shift=shift, diff=diff)
        self.Modelling = Modelling(mode=mode, shuffle=shuffle, store_models=store_models,
                                   scoring=metrics.SCORERS[self.objective],
                                   store_results=False, folder=self.mainDir + 'Models/')

        # Store production
        self.bestModel = None
        self.bestFeatures = None
        self.bestScaler = None
        self.bestOScaler = None

    def _set_flags(self):
        if self.plotEDA is None:
            self.plotEDA = Utils.boolean_input('Make all EDA graphs?')
        if self.processData is None:
            self.processData = Utils.boolean_input('Process/prepare data?')
        if self.validateResults is None:
            self.validateResults = Utils.boolean_input('Validate results?')

    def _load_version(self):
        if self.version is None:
            versions = os.listdir(self.mainDir + 'Production')
            # Updates changelog
            if self.processData:
                if len(versions) == 0:
                    if self.verbose > 0:
                        print('[AutoML] No Production files found. Setting version 0.')
                    self.version = 0
                    file = open(self.mainDir + 'changelog.txt', 'w')
                    file.write('Dataset changelog. \nv0: Initial')
                    file.close()
                else:
                    self.version = len((versions))

                    # Check if not already started
                    with open(self.mainDir + 'changelog.txt', 'r') as f:
                        changelog = f.read()

                    # Else ask for changelog
                    if 'v%i' % self.version in changelog:
                        changelog = changelog[changelog.find('v%i' % self.version):]
                        changelog = changelog[:max(0, changelog.find('\n'))]
                    else:
                        changelog = '\nv%i: ' % self.version + input("Data changelog v%i:\n" % self.version)
                        file = open(self.mainDir + 'changelog.txt', 'a')
                        file.write(changelog)
                        file.close()
                    if self.verbose > 0:
                        print('[AutoML] Set version %s' % (changelog[1:]))
            else:
                if len(versions) == 0:
                    if self.verbose > 0:
                        print('[AutoML] No Production files found. Setting version 0.')
                    self.version = 0
                else:
                    self.version = int(len(versions)) - 1
                    with open(self.mainDir + 'changelog.txt', 'r') as f:
                        changelog = f.read()
                    changelog = changelog[changelog.find('v%i' % self.version):]
                    if self.verbose > 0:
                        print('[AutoML] Loading last version (%s)' % changelog[:changelog.find('\n')])

    def _create_dirs(self):
        dirs = ['', 'Data', 'Features', 'Models', 'Production', 'Validation', 'Sets']
        for dir in dirs:
            try:
                os.makedirs(self.mainDir + dir)
            except:
                continue

    @staticmethod
    def _sort_results(results):
        results['worst_case'] = results['mean_objective'] - results['std_objective']
        return results.sort_values('worst_case', ascending=False)

    def _get_best_params(self, model, feature_set):
        # Filter results for model and version
        results = self.results[np.logical_and(
            self.results['model'] == type(model).__name__,
            self.results['data_version'] == self.version,
        )]

        # Filter results for feature set & sort them
        results = self._sort_results(results[results['dataset'] == feature_set])

        # Warning for unoptimized results
        if 'Hyper Parameter' not in results['type'].values:
            warnings.warn('Hyperparameters not optimized for this combination')

        # Parse & return best parameters (regardless of if it's optimized)
        return Utils.parse_json(results.iloc[0]['params'])

    def fit(self, data):
        '''
        Fit the full autoML pipeline.
        2. (optional) Exploratory Data Analysis
        Creates a ton of plots which are helpful to improve predictions manually
        3. Data Processing
        Cleans all the data. See @DataProcessing
        4. Feature Processing
        Extracts & Selects. See @FeatureProcessing
        5. Initial Modelling
        Runs 12 off the shelf models with default parameters for all feature sets
        If Sequencing is enabled, this is where it happens, as here, the feature set is generated.
        6. Grid Search
        Optimizes the hyperparameters of the best performing models
        7. Prepare Production Files
        Nicely organises all required scripts / files to make a prediction

        @param data: DataFrame including target
        '''
        # Execute pipeline
        self._data_processing(data)
        self._eda()
        self._feature_processing()
        self._initial_modelling()
        self.grid_search()
        # Production Env
        if not os.path.exists(self.mainDir + 'Production/v%i/' % self.version) or \
                len(os.listdir(self.mainDir + 'Production/v%i/' % self.version)) == 0:
            self._prepare_production_files()
        print('[AutoML] Done :)')

    def _eda(self):
        if self.plotEDA:
            print('[AutoML] Starting Exploratory Data Analysis')
            self.eda = DataExploring(self.X, Y=self.Y, folder=self.mainDir, version=self.version)

    def _data_processing(self, data):
        # Load if possible
        if os.path.exists(self.mainDir + 'Data/Cleaned_v%i.csv' % self.version):
            print('[AutoML] Loading Cleaned Data')
            data = pd.read_csv(self.mainDir + 'Data/Cleaned_v%i.csv' % self.version, index_col='index')

        # Clean
        else:
            print('[AutoML] Cleaning Data')
            data = self.DataProcessing.clean(data)

        # Split and store in memory
        self.Y = data[[self.target]]
        self.X = data
        if self.includeOutput is False:
            self.X = self.X.drop(self.target, axis=1)

        # Assert classes in case of classification
        if self.mode == 'classification':
            if self.Y[self.target].nunique() >= 50:
                warnings.warn('More than 50 classes, you might want to reconsider')
            if set(self.Y[self.target]) != set([i for i in range(len(set(self.Y[self.target])))]):
                warnings.warn('Classes should be [0, 1, ...]')

    def _feature_processing(self):
        # Extract
        self.X = self.FeatureProcessing.extract(self.X, self.Y[self.target])

        # Select
        self.colKeep = self.FeatureProcessing.select(self.X, self.Y[self.target])

    def _initial_modelling(self):
        # Load existing results
        if 'Results.csv' in os.listdir(self.mainDir):
            self.results = pd.read_csv(self.mainDir + 'Results.csv')
            self.Modelling.samples = len(self.Y)

        # Check if this version has been modelled
        if self.results is not None and \
                (self.version == self.results['data_version']).any():
            self.results = self._sort_results(self.results)

        # Run Modelling
        else:
            for feature_set, cols in self.colKeep.items():
                # Skip empty sets
                if len(cols) == 0:
                    print('[AutoML] Skipping %s features, empty set' % feature_set)
                else:
                    print('[AutoML] Initial Modelling for %s features (%i)' % (feature_set, len(cols)))

                    # Apply Feature Set
                    self.Modelling.dataset = feature_set
                    # X, Y = self.X.reindex(columns=cols), self.Y.loc[:, self.target]
                    X, Y = self.X[cols], self.Y

                    # Normalize Feature Set (Done here to get one normalization file per feature set)
                    if self.normalize:
                        normalizeFeatures = [k for k in X.keys() if k not in self.dateCols + self.catCols]
                        scaler = StandardScaler()
                        X[normalizeFeatures] = scaler.fit_transform(X[normalizeFeatures])
                        pickle.dump(scaler,
                                    open(self.mainDir + 'Features/Scaler_%s_%i.pickle' % (feature_set, self.version),
                                         'wb'))
                        if self.mode == 'regression':
                            oScaler = StandardScaler()
                            Y[self.target] = oScaler.fit_transform(Y)
                            pickle.dump(oScaler,
                                        open(self.mainDir + 'Features/OScaler_%s_%i.pickle' % (
                                            feature_set, self.version),
                                             'wb'))

                    # Sequence if necessary
                    if self.sequence:
                        X, Y = self.Sequence.convert(X, Y)

                    # Do the modelling
                    results = self.Modelling.fit(X, Y)

                    # Add results to memory
                    results['type'] = 'Initial modelling'
                    results['data_version'] = self.version
                    if self.results is None:
                        self.results = results
                    else:
                        self.results = self.results.append(results)

            # Save results
            self.results.to_csv(self.mainDir + 'Results.csv', index=False)

    def _get_hyper_params(self, model):
        # todo Integrate optuna, baseGridSearch
        # todo parameterize
        # todo change if statements all to model.__module__
        # Parameters for both Regression / Classification
        if isinstance(model, sklearn.linear_model.Lasso) or \
                isinstance(model, sklearn.linear_model.Ridge) or \
                isinstance(model, sklearn.linear_model.RidgeClassifier):
            return {
                'alpha': uniform(0, 10),
            }
        elif isinstance(model, sklearn.svm.SVC) or \
                isinstance(model, sklearn.svm.SVR):
            return {
                'gamma': ['scale', 'auto', 0.001, 0.01, 0.1, 0.5, 1],
                'C': uniform(0, 10),
            }
        elif isinstance(model, sklearn.neighbors.KNeighborsRegressor) or \
                isinstance(model, sklearn.neighbors.KNeighborsClassifier):
            return {
                'n_neighbors': randint(5, 50),
                'weights': ['uniform', 'distance'],
                'leaf_size': randint(10, 150),
                'n_jobs': [mp.cpu_count() - 1],
            }
        elif isinstance(model, sklearn.neural_network._multilayer_perceptron.MLPClassifier) or \
                isinstance(model, sklearn.neural_network._multilayer_perceptron.MLPRegressor):
            return {
                'hidden_layer_sizes': [(100,), (100, 100), (100, 50), (200, 200), (200, 100), (200, 50),
                                       (50, 50, 50, 50)],
                'learning_rate': ['adaptive', 'invscaling'],
                'alpha': [1e-4, 1e-3, 1e-5],
                'shuffle': [False],
            }

        # Regressor specific hyperparameters
        elif self.mode == 'regression':
            if isinstance(model, sklearn.linear_model.SGDRegressor):
                return {
                    'loss': ['squared_loss', 'huber', 'epsilon_insensitive', 'squared_epsilon_insensitive'],
                    'penalty': ['l2', 'l1', 'elasticnet'],
                    'alpha': randint(0, 5),
                }
            elif isinstance(model, sklearn.tree.DecisionTreeRegressor):
                return {
                    'criterion': ['mse', 'friedman_mse', 'mae', 'poisson'],
                    'max_depth': randint(5, 50),
                }
            elif isinstance(model, sklearn.ensemble.AdaBoostRegressor):
                return {
                    'n_estimators': randint(25, 250),
                    'loss': ['linear', 'square', 'exponential'],
                    'learning_rate': uniform(0, 1)
                }
            elif isinstance(model, sklearn.ensemble.GradientBoostingRegressor):
                return {
                    'loss': ['ls', 'lad', 'huber'],
                    'learning_rate': uniform(0, 1),
                    'max_depth': randint(3, 15),
                }
            elif isinstance(model, sklearn.ensemble.HistGradientBoostingRegressor):
                return {
                    'max_iter': randint(100, 250),
                    'max_bins': randint(100, 255),
                    'loss': ['least_squares', 'least_absolute_deviation'],
                    'l2_regularization': uniform(0, 10),
                    'learning_rate': uniform(0, 1),
                    'max_leaf_nodes': randint(30, 150),
                    'early_stopping': [True],
                }
            elif isinstance(model, sklearn.ensemble.RandomForestRegressor):
                return {
                    'criterion': ['mse', 'mae'],
                    'max_depth': randint(3, 15),
                    'max_features': ['auto', 'sqrt'],
                    'min_samples_split': randint(2, 50),
                    'min_samples_leaf': randint(1, 1000),
                    'bootstrap': [True, False],
                }
            elif model.__module__ == 'catboost.core':
                return {
                    'loss_function': ['MAE', 'RMSE'],
                    'learning_rate': uniform(0, 1),
                    'l2_leaf_reg': uniform(0, 10),
                    'depth': randint(3, 15),
                    'min_data_in_leaf': randint(1, 1000),
                    'max_leaves': randint(10, 250),
                }
            elif model.__module__ == 'xgboost.sklearn':
                return {
                    'max_depth': randint(3, 15),
                    'booster': ['gbtree', 'gblinear', 'dart'],
                    'learning_rate': uniform(0, 10),
                    'verbosity': [0],
                    'n_jobs': [mp.cpu_count() - 1],
                }
            elif model.__module__ == 'lightgbm.sklearn':
                return {
                    'num_leaves': randint(10, 150),
                    'min_child_samples': randint(1, 1000),
                    'min_child_weight': uniform(0, 1),
                    'subsample': uniform(0, 1),
                    'colsample_bytree': uniform(0, 1),
                    'reg_alpha': uniform(0, 1),
                    'reg_lambda': uniform(0, 1),
                    'n_jobs': [mp.cpu_count() - 1],
                }

        # Classification specific hyperparameters
        elif self.mode == 'classification':
            if isinstance(model, sklearn.linear_model.SGDClassifier):
                return {
                    'loss': ['hinge', 'log', 'modified_huber', 'squared_hinge'],
                    'penalty': ['l2', 'l1', 'elasticnet'],
                    'alpha': uniform(0, 10),
                    'max_iter': randint(250, 2000),
                }
            elif isinstance(model, sklearn.tree.DecisionTreeClassifier):
                return {
                    'criterion': ['gini', 'entropy'],
                    'max_depth': randint(5, 50),
                }
            elif isinstance(model, sklearn.ensemble.AdaBoostClassifier):
                return {
                    'n_estimators': randint(25, 250),
                    'learning_rate': uniform(0, 1)
                }
            elif isinstance(model, sklearn.ensemble.BaggingClassifier):
                return {
                    # 'n_estimators': [5, 10, 15, 25, 50],
                    'max_features': uniform(0, 1),
                    'bootstrap': [False, True],
                    'bootstrap_features': [True, False],
                    'n_jobs': [mp.cpu_count() - 1],
                }
            elif isinstance(model, sklearn.ensemble.GradientBoostingClassifier):
                return {
                    'loss': ['deviance', 'exponential'],
                    'learning_rate': uniform(0, 1),
                    'max_depth': randint(3, 15),
                }
            elif isinstance(model, sklearn.ensemble.HistGradientBoostingClassifier):
                return {
                    'max_iter': randint(100, 250),
                    'max_bins': randint(100, 255),
                    'l2_regularization': uniform(0, 10),
                    'learning_rate': uniform(0, 1),
                    'max_leaf_nodes': randint(30, 150),
                    'early_stopping': [True]
                }
            elif isinstance(model, sklearn.ensemble.RandomForestClassifier):
                return {
                    'criterion': ['gini', 'entropy'],
                    'max_depth': randint(3, 15),
                    'max_features': ['auto', 'sqrt'],
                    'min_samples_split': randint(2, 50),
                    'min_samples_leaf': randint(1, 1000),
                    'bootstrap': [True, False],
                }
            elif model.__module__ == 'catboost.core':
                return {
                    'loss_function': ['Logloss' if self.Y[self.target].nunique() == 2 else 'MultiClass'],
                    'learning_rate': uniform(0, 1),
                    'l2_leaf_reg': uniform(0, 10),
                    'depth': randint(1, 10),
                    'min_data_in_leaf': randint(50, 500),
                    'grow_policy': ['SymmetricTree', 'Depthwise', 'Lossguide'],
                }
            elif model.__module__ == 'xgboost.sklearn':
                return {
                    'max_depth': randint(3, 15),
                    'booster': ['gbtree', 'gblinear', 'dart'],
                    'learning_rate': uniform(0, 10),
                    'verbosity': [0],
                    'n_jobs': [mp.cpu_count() - 1],
                    'scale_pos_weight': uniform(0, 100)
                }
            elif model.__module__ == 'lightgbm.sklearn':
                return {
                    'num_leaves': randint(10, 150),
                    'min_child_samples': randint(1, 1000),
                    'min_child_weight': uniform(0, 1),
                    'subsample': uniform(0, 1),
                    'colsample_bytree': uniform(0, 1),
                    'reg_alpha': uniform(0, 1),
                    'reg_lambda': uniform(0, 1),
                    'n_jobs': [mp.cpu_count() - 1],
                }

        # Raise error if nothing is returned
        raise NotImplementedError('Hyperparameter tuning not implemented for ', type(model).__name__)

    def grid_search(self, model=None, feature_set=None, parameter_set=None):
        """
        Runs a grid search. By default, takes the self.results, and runs for the top 3 optimizations.
        There is the option to provide a model & feature_set, but both have to be provided. In this case,
        the model & data set combination will be optimized.
        Implemented types, Base, Halving, Optuna
        """
        assert model is not None and feature_set is not None or model == feature_set, \
            'Model & feature_set need to be either both None or both provided.'

        # If arguments are provided
        if model is not None:

            # Get model string
            if isinstance(model, str):
                models = self.Modelling.return_models()
                model = models[[i for i in range(len(models)) if type(models[i]).__name__ == model][0]]

            # Organise existing results
            results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['data_version'] == self.version,
            )]
            results = self._sort_results(results[results['dataset'] == feature_set])

            # Check if exists and load
            if ('Hyper Parameter' == results['type']).any():
                print('[AutoML] Loading optimization results.')
                hyper_opt_results = results[results['type'] == 'Hyper Parameter']
                params = Utils.parse_json(hyper_opt_results.iloc[0]['params'])

            # Or run
            else:
                # Parameter check
                if parameter_set is None:
                    parameter_set = self._get_hyper_params(model)
                    
                # Run grid search
                grid_search_results = self._sort_results(self._grid_search_iteration(model, parameter_set, feature_set))

                # Store results
                grid_search_results['model'] = type(model).__name__
                grid_search_results['data_version'] = self.version
                grid_search_results['dataset'] = feature_set 
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)

                # Get params for validation
                params = Utils.parse_json(grid_search_results.iloc[0]['params'])

            # Validate
            self._validate_result(model, params, feature_set)
            return

        # If arguments aren't provided, run through promising models
        models = self.Modelling.return_models()
        results = self._sort_results(self.results[np.logical_and(
            self.results['type'] == 'Initial modelling',
            self.results['data_version'] == self.version,
        )])
        for iteration in range(self.gridSearchIterations):
            # Grab settings
            settings = results.iloc[iteration]
            model = models[[i for i in range(len(models)) if type(models[i]).__name__ == settings['model']][0]]
            feature_set = settings['dataset']

            # Check whether exists
            model_results = self.results[np.logical_and(
                self.results['model'] == type(model).__name__,
                self.results['data_version'] == self.version,
            )]
            model_results = self._sort_results(model_results[model_results['dataset'] == feature_set])

            # If exists
            if ('Hyper Parameter' == model_results['type']).any():
                hyper_opt_res = model_results[model_results['type'] == 'Hyper Parameter']
                params = Utils.parse_json(hyper_opt_res.iloc[0]['params'])

            # Else run
            else:
                # Get params
                params = self._get_hyper_params(model)

                grid_search_results = self._sort_results(self._grid_search_iteration(model, params, feature_set))

                # Store
                grid_search_results['model'] = type(model).__name__
                grid_search_results['data_version'] = self.version
                grid_search_results['dataset'] = feature_set
                grid_search_results['type'] = 'Hyper Parameter'
                self.results = self.results.append(grid_search_results)
                self.results.to_csv(self.mainDir + 'Results.csv', index=False)
                params = Utils.parseJson(grid_search_results.iloc[0]['params'])

            # Validate
            if self.validateResults:
                self._validate_result(model, params, feature_set)

    def _grid_search_iteration(self, model, params, feature_set):
        """
        INTERNAL | Grid search for defined model, parameter set and feature set.
        """
        print('\n[AutoML] Starting Hyper Parameter Optimization for %s on %s features (%i samples, %i features)' %
              (type(model).__name__, feature_set, len(self.X), len(self.colKeep[feature_set])))

        # Select data
        x, y = self.X[self.colKeep[feature_set]], self.Y

        # Normalize Feature Set (the input remains original)
        if self.normalize:
            features_to_normalize = [k for k in self.colKeep[feature_set] if k not in self.dateCols + self.catCols]
            scaler = pickle.load(
                open(self.mainDir + 'Features/Scaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
            x[features_to_normalize] = scaler.transform(x[features_to_normalize])
            if self.mode == 'regression':
                output_scaler = pickle.load(
                    open(self.mainDir + 'Features/OScaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
                y = output_scaler.transform(y)

        # Cross-Validator
        if self.mode == 'regression':
            cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)
        elif self.mode == 'classification':
            cv = StratifiedKFold(n_splits=self.cvSplits, shuffle=self.shuffle)

        # Select right hyperparameter optimizer
        if self.gridSearchType == 'Base':
            grid_search = BaseGridSearch(model, params=params, cv=cv, scoring=self.scorer, verbose=self.verbose)
        elif self.gridSearchType == 'Halving':
            grid_search = HalvingGridSearch(model, params=params, cv=cv, scoring=self.scorer, verbose=self.verbose)
        elif self.gridSearchType == 'Optuna':
            grid_search = OptunaGridSearch(model, params=params, cv=cv, scoring=self.scorer, verbose=self.verbose)
        else:
            raise NotImplementedError('Only Base, Halving and Optuna are implemented.')
        # Get results
        results = grid_search.fit(x, y)
        results['worst_case'] = results['mean_objective'] - results['std_objective']
        return results.sort_values('worst_case', ascending=False)

    def _create_stacking(self):
        # todo implement
        '''
        Based on the best performing models, in addition to cheap models based on very different assumptions,
        A stacking model is optimized to enhance/combine the performance of the models.
        '''
        # First, the ideal dataset has to be chosen, we're restricted to a single one...
        results = self._sort_results(self.results[np.logical_and(
            self.results['type'] == 'Hyper Parameter',
            self.results['data_version'] == self.version,
        )])
        feature_set = results['dataset'].value_counts()[0]

    def validate(self, model, feature_set, params=None):
        '''
        Just a wrapper for the outside.
        Parameters:
        Model: The model to optimize, either string or class
        Feature Set: String
        (optional) params: Model parameters for which to validate
        '''
        assert feature_set in self.colKeep.keys(), 'Feature Set not available.'

        # Get model
        if isinstance(model, str):
            models = self.Modelling.return_models()
            model = models[[i for i in range(len(models)) if type(models[i]).__name__ == model][0]]

        # Get params
        if params is not None:
            params = self._getBestParams(model, feature_set)

        # Run validation
        self._validateResult(model, params, feature_set)

    def _validate_result(self, master_model, params, feature_set):
        print('[AutoML] Validating results for %s (%i %s features) (%s)' % (type(master_model).__name__,
                                                                            len(self.colKeep[feature_set]), feature_set,
                                                                            params))
        if not os.path.exists(self.mainDir + 'Validation/'): os.mkdir(self.mainDir + 'Validation/')

        # Select data
        X, Y = self.X[self.colKeep[feature_set]], self.Y

        # Normalize Feature Set (the X remains original)
        if self.normalize:
            normalizeFeatures = [k for k in self.colKeep[feature_set] if k not in self.dateCols + self.catCols]
            scaler = pickle.load(
                open(self.mainDir + 'Features/Scaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
            X[normalizeFeatures] = scaler.transform(X[normalizeFeatures])
            if self.mode == 'regression':
                oScaler = pickle.load(
                    open(self.mainDir + 'Features/OScaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
                Y[Y.keys()] = oScaler.transform(Y)
                print('(%.1f, %.1f)' % (np.mean(Y), np.std(Y)))
        X, Y = X.to_numpy(), Y.to_numpy().reshape((-1, 1))

        # For Regression
        if self.mode == 'regression':

            # Cross-Validation Plots
            fig, ax = plt.subplots(math.ceil(self.cvSplits / 2), 2, sharex=True, sharey=True)
            fig.suptitle('%i-Fold Cross Validated Predictions - %s' % (self.cvSplits, type(master_model).__name__))

            # Initialize iterables
            score = []
            cv = KFold(n_splits=self.cvSplits, shuffle=self.shuffle)
            # Cross Validate
            for i, (t, v) in enumerate(cv.split(X, Y)):
                Xt, Xv, Yt, Yv = X[t], X[v], Y[t].reshape((-1)), Y[v].reshape((-1))
                model = copy.copy(master_model)
                model.set_params(**params)
                model.fit(Xt, Yt)

                # Metrics
                score.append(self.scorer(model, Xv, Yv))

                # Plot
                ax[i // 2][i % 2].set_title('Fold-%i' % i)
                ax[i // 2][i % 2].plot(Yv, color='#2369ec')
                ax[i // 2][i % 2].plot(model.predict(Xv), color='#ffa62b', alpha=0.4)

            # Print & Finish plot
            print('[AutoML] %s:        %.2f \u00B1 %.2f' % (self.scorer._score_func.__name__,
                                                            np.mean(score), np.std(score)))
            ax[i // 2][i % 2].legend(['Output', 'Prediction'])
            plt.show()

        # For BINARY classification
        elif self.mode == 'classification' and self.Y[self.target].nunique() == 2:
            # Initiating
            fig, ax = plt.subplots(math.ceil(self.cvSplits / 2), 2, sharex=True, sharey=True)
            fig.suptitle('%i-Fold Cross Validated Predictions - %s (%s)' %
                         (self.cvSplits, type(master_model).__name__, feature_set))
            acc = []
            prec = []
            rec = []
            spec = []
            f1 = []
            aucs = []
            tprs = []
            cm = np.zeros((2, 2))
            mean_fpr = np.linspace(0, 1, 100)

            # Modelling
            cv = StratifiedKFold(n_splits=self.cvSplits)
            for i, (t, v) in enumerate(cv.split(X, Y)):
                n = len(v)
                Xt, Xv, Yt, Yv = X[t], X[v], Y[t].reshape((-1)), Y[v].reshape((-1))
                model = copy.copy(master_model)
                model.set_params(**params)
                model.fit(Xt, Yt)
                predictions = model.predict(Xv).reshape((-1))

                # Metrics
                tp = np.logical_and(np.sign(predictions) == 1, Yv == 1).sum()
                tn = np.logical_and(np.sign(predictions) == 0, Yv == 0).sum()
                fp = np.logical_and(np.sign(predictions) == 1, Yv == 0).sum()
                fn = np.logical_and(np.sign(predictions) == 0, Yv == 1).sum()
                acc.append((tp + tn) / n * 100)
                if tp + fp > 0:
                    prec.append(tp / (tp + fp) * 100)
                if tp + fn > 0:
                    rec.append(tp / (tp + fn) * 100)
                if tn + fp > 0:
                    spec.append(tn / (tn + fp) * 100)
                if tp + fp > 0 and tp + fn > 0:
                    f1.append(2 * prec[-1] * rec[-1] / (prec[-1] + rec[-1]) if prec[-1] + rec[-1] > 0 else 0)
                cm += np.array([[tp, fp], [fn, tn]]) / self.cvSplits

                # Plot
                ax[i // 2][i % 2].plot(Yv, c='#2369ec', alpha=0.6)
                ax[i // 2][i % 2].plot(predictions, c='#ffa62b')
                ax[i // 2][i % 2].set_title('Fold-%i' % i)

            # Results
            print('[AutoML] Accuracy:        %.2f \u00B1 %.2f %%' % (np.mean(acc), np.std(acc)))
            print('[AutoML] Precision:       %.2f \u00B1 %.2f %%' % (np.mean(prec), np.std(prec)))
            print('[AutoML] Recall:          %.2f \u00B1 %.2f %%' % (np.mean(rec), np.std(rec)))
            print('[AutoML] Specificity:     %.2f \u00B1 %.2f %%' % (np.mean(spec), np.std(spec)))
            print('[AutoML] F1-score:        %.2f \u00B1 %.2f %%' % (np.mean(f1), np.std(f1)))
            print('[AutoML] Confusion Matrix:')
            print('[AutoML] Pred \ true |  Faulty   |   Healthy      ')
            print('[AutoML]  Faulty     |  %s|  %.1f' % (('%.1f' % cm[0, 0]).ljust(9), cm[0, 1]))
            print('[AutoML]  Healthy    |  %s|  %.1f' % (('%.1f' % cm[1, 0]).ljust(9), cm[1, 1]))

            # Check whether plot is possible
            if isinstance(model, sklearn.linear_model.Lasso) or isinstance(model, sklearn.linear_model.Ridge):
                return

            # Plot ROC
            fig, ax = plt.subplots(figsize=[12, 8])
            viz = metrics.plot_roc_curve(model, Xv, Yv, name='ROC fold {}'.format(i + 1), alpha=0.3, ax=ax)
            interp_tpr = np.interp(mean_fpr, viz.fpr, viz.tpr)
            interp_tpr[0] = 0.0
            tprs.append(interp_tpr)
            aucs.append(viz.roc_auc)

            # Adjust plots
            ax.plot([0, 1], [0, 1], linestyle='--', lw=2, color='#ffa62b',
                    label='Chance', alpha=.8)
            mean_tpr = np.mean(tprs, axis=0)
            mean_tpr[-1] = 1.0
            mean_auc = metrics.auc(mean_fpr, mean_tpr)
            std_auc = np.std(aucs)
            ax.plot(mean_fpr, mean_tpr, color='#2369ec',
                    label=r'Mean ROC (AUC = %0.2f $\pm$ %0.2f)' % (mean_auc, std_auc),
                    lw=2, alpha=.8)
            std_tpr = np.std(tprs, axis=0)
            tprs_upper = np.minimum(mean_tpr + std_tpr, 1)
            tprs_lower = np.maximum(mean_tpr - std_tpr, 0)
            ax.fill_between(mean_fpr, tprs_lower, tprs_upper, color='#729ce9', alpha=.2,
                            label=r'$\pm$ 1 std. dev.')
            ax.set(xlim=[-0.05, 1.05], ylim=[-0.05, 1.05],
                   title="ROC Curve - %s" % type(master_model).__name__)
            ax.legend(loc="lower right")
            fig.savefig(self.mainDir + 'Validation/ROC_%s.png' % type(model).__name__, format='png', dpi=200)

        # For MULTICLASS classification
        elif self.mode == 'classification':
            # Initiating
            fig, ax = plt.subplots(math.ceil(self.cvSplits / 2), 2, sharex=True, sharey=True)
            fig.suptitle('%i-Fold Cross Validated Predictions - %s (%s)' %
                         (self.cvSplits, type(master_model).__name__, feature_set))
            n_classes = self.Y[self.target].nunique()
            f1Score = np.zeros((self.cvSplits, n_classes))
            logLoss = np.zeros(self.cvSplits)
            avgAcc = np.zeros(self.cvSplits)

            # Modelling
            cv = StratifiedKFold(n_splits=self.cvSplits)
            for i, (t, v) in enumerate(cv.split(X, Y)):
                n = len(v)
                Xt, Xv, Yt, Yv = X[t], X[v], Y[t].reshape((-1)), Y[v].reshape((-1))
                model = copy.copy(master_model)
                model.set_params(**params)
                model.fit(Xt, Yt)
                predictions = model.predict(Xv).reshape((-1))

                # Metrics
                f1Score[i] = metrics.f1_score(Yv, predictions, average=None)
                avgAcc[i] = metrics.accuracy_score(Yv, predictions)
                if hasattr(model, 'predict_proba'):
                    probabilities = model.predict_proba(Xv)
                    logLoss[i] = metrics.log_loss(Yv, probabilities)

                # Plot
                ax[i // 2][i % 2].plot(Yv, c='#2369ec', alpha=0.6)
                ax[i // 2][i % 2].plot(predictions, c='#ffa62b')
                ax[i // 2][i % 2].set_title('Fold-%i' % i)

            # Results
            print('F1 scores:')
            print(''.join([' Class %i |' % i for i in range(n_classes)]))
            print(''.join([' %.2f '.ljust(9) % f1 + '|' for f1 in np.mean(f1Score, axis=0)]))
            print('Average Accuracy: %.2f \u00B1 %.2f' % (np.mean(avgAcc), np.std(avgAcc)))
            if hasattr(model, 'predict_proba'):
                print('Log Loss:         %.2f \u00B1 %.2f' % (np.mean(logLoss), np.std(logLoss)))

    def _prepare_production_files(self, model=None, feature_set=None, params=None):
        if not os.path.exists(self.mainDir + 'Production/v%i/' % self.version):
            os.mkdir(self.mainDir + 'Production/v%i/' % self.version)
        # Get sorted results for this data version
        results = self._sort_results(self.results[self.results['data_version'] == self.version])

        # In the case args are provided
        if model is not None and feature_set is not None:
            # Take name if model instance is given
            if not isinstance(model, str):
                model = type(model).__name__
            if params is None:
                results = self._sortResults(
                    results[np.logical_and(results['model'] == model, results['dataset'] == feature_set)])
                params = Utils.parseJson(results.iloc[0]['params'])

        # Otherwise find best
        else:
            model = results.iloc[0]['model']
            feature_set = results.iloc[0]['dataset']
            params = results.iloc[0]['params']
            if isinstance(params, str):
                params = Utils.parseJson(params)

        # Notify of results
        print('[AutoML] Preparing Production Env Files for %s, feature set %s' %
              (model, feature_set))
        print('[AutoML] ', params)
        print('[AutoML] %s: %.2f \u00B1 %.2f' %
              (self.scorer._score_func.__name__, results.iloc[0]['mean_objective'], results.iloc[0]['std_objective']))

        # Save Features
        self.bestFeatures = self.colKeep[feature_set]
        json.dump(self.bestFeatures, open(self.mainDir + 'Production/v%i/Features.json' % self.version, 'w'))

        # Copy data
        X, Y = self.X[self.bestFeatures], self.Y

        # Save Scalers & Normalize
        if self.normalize:
            # Save
            shutil.copy(self.mainDir + 'Features/Scaler_%s_%i.pickle' % (feature_set, self.version),
                        self.mainDir + 'Production/v%i/Scaler.pickle' % self.version)
            if self.mode == 'regression':
                shutil.copy(self.mainDir + 'Features/OScaler_%s_%i.pickle' % (feature_set, self.version),
                            self.mainDir + 'Production/v%i/OScaler.pickle' % self.version)

            # Normalize
            normalizeFeatures = [k for k in self.bestFeatures if k not in self.dateCols + self.catCols]
            self.bestScaler = pickle.load(
                open(self.mainDir + 'Features/Scaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
            X = self.bestScaler.transform(X[normalizeFeatures])
            if self.mode == 'regression':
                self.bestOScaler = pickle.load(
                    open(self.mainDir + 'Features/OScaler_%s_%i.pickle' % (feature_set, self.version), 'rb'))
                Y = self.bestOScaler.transform(Y)
            else:
                Y = self.Y[self.target].to_numpy()

        # Cluster Features require additional 'Centers' file
        if any(['dist__' in key for key in self.bestFeatures]):
            shutil.copy(self.mainDir + 'Features/KMeans_v%i.csv' % self.version,
                        self.mainDir + 'Production/v%i/KMeans.csv' % self.version)

        # Model
        self.bestModel = [mod for mod in self.Modelling.return_models() if type(mod).__name__ == model][0]
        self.bestModel.set_params(**params)
        self.bestModel.fit(X, Y.values.ravel())
        joblib.dump(self.bestModel, self.mainDir + 'Production/v%i/Model.joblib' % self.version)

        # Predict function
        predictCode = self.create_predict_function(self.customCode)
        with open(self.mainDir + 'Production/v%i/Predict.py' % self.version, 'w') as f:
            f.write(predictCode)
        with open(self.mainDir + 'Production/v%i/__init__.py' % self.version, 'w') as f:
            f.write('')

        # Pipeline
        pickle.dump(self, open(self.mainDir + 'Production/v%i/Pipeline.pickle' % self.version, 'wb'))
        return

    def _error_analysis(self):
        # todo implement
        pass

    def _convert_data(self, data):
        # Load files
        folder = 'Production/v%i/' % self.version
        features = json.load(open(self.mainDir + folder + 'Features.json', 'r'))

        if self.normalize:
            scaler = pickle.load(open(self.mainDir + folder + 'Scaler.pickle', 'rb'))
            if self.mode == 'regression':
                oScaler = pickle.load(open(self.mainDir + folder + 'OScaler.pickle', 'rb'))
        if self.mode == 'regression':
            oScaler = pickle.load(open(self.mainDir + folder + 'OScaler.pickle', 'rb'))

        # Clean data
        # todo change to transform
        data = Utils.clean_keys(data)
        data = self.DataProcessing._convert_data_types(data)
        data = self.DataProcessing._remove_duplicates(data)
        data = self.DataProcessing._remove_outliers(data)
        data = self.DataProcessing._remove_missing_values(data)
        if data.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
            raise ValueError('Data should not contain NaN or Infs after cleaning!')

        # Save output
        if self.target in data.keys():
            Y = data[self.target].to_numpy().reshape((-1, 1))
        else:
            Y = None

        # Convert Features
        if 'KMeans.csv' in os.listdir(self.mainDir + folder):
            k_means = pd.read_csv(self.mainDir + folder + 'KMeans.csv')
            X = self.FeatureProcessing.transform(data=data, features=features, k_means=k_means)
        else:
            X = self.FeatureProcessing.transform(data, features)

        if X.astype('float32').replace([np.inf, -np.inf], np.nan).isna().sum().sum() != 0:
            raise ValueError('Data should not contain NaN or Infs after adding features!')

        # Normalize
        if self.normalize:
            X[X.keys()] = scaler.transform(X)

        # Return
        return X, Y

    def predict(self, data):
        '''
        Full script to make predictions. Uses 'Production' folder with defined or latest version.
        '''
        # Feature Extraction, Selection and Normalization
        model = joblib.load(self.mainDir + 'Production/v%i/Model.joblib' % self.version)
        if self.verbose > 0:
            print('[AutoML] Predicting with %s, v%i' % (type(model).__name__, self.version))
        X, Y = self._convert_data(data)

        # Predict
        if self.mode == 'regression':
            if self.normalize:
                oScaler = pickle.load(open('Production/v%i/OScaler.pickle', 'rb'))
                predictions = oScaler.inverse_transform(model.predict(X))
            else:
                predictions = model.predict(X)
        if self.mode == 'classification':
            try:
                predictions = model.predict_proba(X)[:, 1]
            except AttributeError:
                predictions = model.predict(X)

        return predictions

    def create_predict_function(self, custom_code):
        '''
        This function returns a string, which can be used to make predictions.
        This is in a predefined format, a Predict class, with a predict funtion taking the arguments
        model: trained sklearn-like class with the .fit() function
        features: list of strings containing all features fed to the model
        scaler: trained sklearn-like class with .transform function
        data: the data to predict on
        Now this function has the arg decoding, which allows custom code injection
        '''
        # Check if predict file exists already to increment version
        if os.path.exists(self.mainDir + 'Production/v%i/Predict.py' % self.version):
            with open(self.mainDir + 'Production/v%i/Predict.py' % self.version, 'r') as f:
                predictFile = f.read()
            ind = predictFile.find('self.version = ') + 16
            oldVersion = predictFile[ind: predictFile.find("'", ind)]
            minorVersion = int(oldVersion[oldVersion.find('.') + 1:])
            version = oldVersion[:oldVersion.find('.') + 1] + str(minorVersion + 1)
        else:
            version = 'v%i.0' % self.version
        print('Creating Prediction %s' % version)
        dataProcess = self.DataProcessing.exportFunction()
        featureProcess = self.FeatureProcessing.exportFunction()
        return """import pandas as pd
import numpy as np
import struct, re, copy, os


class Predict(object):

    def __init__(self):
        self.version = '{}'

    def predict(self, model, features, data, **args):
        ''' 
        Prediction function for Amplo's AutoML. 
        This is in a predefined format: 
        - a 'Predict' class, with a 'predict' funtion taking the arguments:        
            model: trained sklearn-like class with the .fit() function
            features: list of strings containing all features fed to the model
            data: the data to predict on
        Note: May depend on additional named arguments within args. 
        '''
        ###############
        # Custom Code #
        ###############""".format(version) + textwrap.indent(custom_code, '    ') \
               + dataProcess + featureProcess + '''
        ###########
        # Predict #
        ###########
        mode, normalize = '{}', {}

        # Normalize
        if normalize:
            assert 'scaler' in args.keys(), 'When Normalizing=True, scaler needs to be provided in args'
            X = args['scaler'].transform(X)

        # Predict
        if mode == 'regression':
            if normalize:
                assert 'o_scaler' in args.keys(), 'When Normalizing=True, o_scaler needs to be provided in args'
                predictions = args['oScaler'].inverse_transform(model.predict(X))
            else:
                predictions = model.predict(X)
        if mode == 'classification':
            try:
                predictions = model.predict_proba(X)[:, 1]
            except:
                predictions = model.predict(X)

        return predictions
'''.format(self.mode, self.normalize)



import os
import copy
import ppscore
import numpy as np
import pandas as pd
from tqdm import tqdm
import seaborn as sns
import missingno
import matplotlib
import matplotlib.pyplot as plt
from statsmodels.tsa.seasonal import STL
from statsmodels.graphics.tsaplots import plot_acf
from statsmodels.graphics.tsaplots import plot_pacf
from sklearn.ensemble import RandomForestRegressor
from sklearn.ensemble import RandomForestClassifier


class DataExploring:

    def __init__(self, data,
                 y=None,
                 plot_time_plots=True,
                 plot_box_plots=False,
                 plot_missing_values=True,
                 plot_seasonality=False,
                 plot_co_linearity=True,
                 plot_differencing=False,
                 plot_signal_correlations=False,
                 plot_feature_importance=True,
                 plot_scatter_plots=False,
                 differ=0,
                 pre_tag='',
                 max_samples=10000,
                 season_periods=None,
                 lags=60,
                 skip_completed=True,
                 folder='',
                 version=None):
        """
        Doing all the fun EDA in an automated manner :)
        """
        assert isinstance(data, pd.DataFrame)

        # Running booleans
        self.plotTimePlots = plot_time_plots
        self.plotBoxPlots = plot_box_plots
        self.plotMissingValues = plot_missing_values
        self.plotSeasonality = plot_seasonality
        self.plotCoLinearity = plot_co_linearity
        self.plotDifferencing = plot_differencing
        self.plotSignalCorrelations = plot_signal_correlations
        self.plotFeatureImportance = plot_feature_importance
        self.plotScatterPlots = plot_scatter_plots

        # Register data
        self.data = data.astype('float32').fillna(0)
        if y is not None:
            assert isinstance(y, pd.DataFrame) or isinstance(y, pd.Series)
            if isinstance(y, pd.DataFrame):
                y = y[y.keys()[0]]
            self.Y = y.astype('float32').fillna(0)
            if self.Y.nunique() == 2:
                print('[AutoML] Mode set to classification.')
                self.mode = 'classification'
                if set(self.Y.values) != {0, 1}:
                    assert 1 in self.Y.values, 'Ambiguous classes (either {0, 1} or {-1, 1})'
                    self.Y.loc[self.Y.values != 1] = 0
            else:
                print('[AutoML] Mode set to regression.')
                self.mode = 'regression'
        else:
            self.mode = None

        # General settings
        self.seasonPeriods = season_periods
        self.maxSamples = max_samples       # Time series
        self.differ = differ                # Correlations
        self.lags = lags                    # Correlations

        # Storage settings
        self.tag = pre_tag
        self.version = version if version is not None else 0
        self.folder = folder if folder == '' or folder[-1] == '/' else folder + '/'
        self.skip = skip_completed

        # Create Base folder
        if not os.path.exists(self.folder + 'EDA/'):
            os.mkdir(self.folder + 'EDA')
        self.folder += 'EDA/'
        self.run()

    def run(self):
        # Run all functions
        if self.mode == 'classification':
            self._run_classification()
        else:
            self._run_regression()

    def _run_classification(self):
        print('[EDA] Generating Missing Values Plot')
        self.missing_values()
        print('[EDA] Generating Time plots')
        self.time_plots()
        print('[EDA] Generating Box plots')
        self.box_plots()
        if self.Y is not None:
            print('[EDA] Generating SHAP plot')
            self.shap()
            print('[EDA] Generating Feature Ranking Plot')
            self.feature_ranking()
            print('[EDA] Predictive Power Score Plot')
            self.predictive_power_score()

    def _run_regression(self):
        print('[EDA] Generating Missing Values Plot')
        self.missing_values()
        print('[EDA] Generating Time plots')
        self.time_plots()
        print('[EDA] Generating Box plots')
        self.box_plots()
        self.seasonality()
        print('[EDA] Generating Co-linearity Plots')
        self.co_linearity()
        print('[EDA] Generating Diff Var Plot')
        self.differencing()
        print('[EDA] Generating ACF Plots')
        self.complete_auto_corr()
        print('[EDA] Generating PACF Plots')
        self.partial_auto_corr()
        if self.Y is not None:
            print('[EDA] Generating CCF Plots')
            self.cross_corr()
            print('[EDA] Generating Scatter plots')
            self.scatters()
            print('[EDA] Generating SHAP plot')
            self.shap()
            print('[EDA] Generating Feature Ranking Plot')
            self.feature_ranking()
            print('[EDA] Predictive Power Score Plot')
            self.predictive_power_score()

    def missing_values(self):
        if self.plotMissingValues:
            # Create folder
            if not os.path.exists(self.folder + 'MissingValues/'):
                os.mkdir(self.folder + 'MissingValues/')

            # Skip if exists
            if self.tag + 'v%i.png' % self.version in os.listdir(self.folder + 'MissingValues/'):
                return

            # Plot
            ax = missingno.matrix(self.data, figsize=[24, 16])
            fig = ax.get_figure()
            fig.savefig(self.folder + 'MissingValues/v%i.png' % self.version)

    def box_plots(self):
        if self.plotBoxPlots:
            # Create folder
            if not os.path.exists(self.folder + 'BoxPlots/v%i/' % self.version):
                os.makedirs(self.folder + 'BoxPlots/v%i/' % self.version)

            # Iterate through vars
            for key in tqdm(self.data.keys()):

                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'BoxPlots/v%i/' % self.version):
                    continue

                # Figure prep
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Classification
                if self.mode == 'classification':
                    plt.boxplot([self.data.loc[self.Y == 1, key], self.data.loc[self.Y == -1, key]], labels=['Faulty',
                                                                                                             'Healthy'])
                    plt.legend(['Faulty', 'Healthy'])

                # Regression
                if self.mode == 'regression':
                    plt.boxplot(self.data[key])

                # Store & Close
                fig.savefig(self.folder + 'BoxPlots/v%i/' % self.version + self.tag + key + '.png',
                            format='png', dpi=300)
                plt.close()

    def time_plots(self):
        if self.plotTimePlots:
            # Create folder
            if not os.path.exists(self.folder + 'TimePlots/v%i/' % self.version):
                os.makedirs(self.folder + 'TimePlots/v%i/' % self.version)

            # Set matplot limit
            matplotlib.use('Agg')
            matplotlib.rcParams['agg.path.chunksize'] = 200000

            # Undersample
            ind = np.linspace(0, len(self.data) - 1, self.maxSamples).astype('int')
            data, y = self.data.iloc[ind], self.Y.iloc[ind]

            # Iterate through features
            for key in tqdm(data.keys()):
                # Skip if existing
                if self.tag + key + '.png' in os.listdir(self.folder + 'TimePlots/v%i/' % self.version):
                    continue

                # Figure preparation
                fig = plt.figure(figsize=[24, 16])
                plt.title(key)

                # Plot
                if self.mode == 'classification':
                    cm = plt.get_cmap('bwr')
                else:
                    cm = plt.get_cmap('summer')
                nm_output = (y - y.min()) / (y.max() - y.min())
                plt.scatter(data.index, data[key], c=cm(nm_output), alpha=0.3)

                # Store & Close
                fig.savefig(self.folder + 'TimePlots/v%i/' % self.version + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def seasonality(self):
        if self.plotSeasonality:
            # Create folder
            if not os.path.exists(self.folder + 'Seasonality/'):
                os.mkdir(self.folder + 'Seasonality/')

            # Iterate through features
            for key in tqdm(self.data.keys()):
                for period in self.seasonPeriods:
                    if self.tag + key + '_v%i.png' % self.version in os.listdir(self.folder + 'Seasonality/'):
                        continue
                    seasonality = STL(self.data[key], period=period).fit()
                    fig = plt.figure(figsize=[24, 16])
                    plt.plot(range(len(self.data)), self.data[key])
                    plt.plot(range(len(self.data)), seasonality)
                    plt.title(key + ', period=' + str(period))
                    fig.savefig(self.folder + 'Seasonality/' + self.tag + str(period)+'/'+key +
                                '_v%i.png' % self.version, format='png', dpi=300)
                    plt.close()

    def co_linearity(self):
        if self.plotCoLinearity:
            # Create folder
            if not os.path.exists(self.folder + 'CoLinearity/v%i/'):
                os.makedirs(self.folder + 'CoLinearity/v%i/')

            # Skip if existing
            if self.tag + 'MinimumRepresentation.png' in os.listdir(self.folder + 'Colinearity/v%i/' % self.version):
                return

            # Plot threshold matrix
            threshold = 0.95
            fig = plt.figure(figsize=[24, 16])
            plt.title('Co-linearity matrix, threshold %.2f' % threshold)
            sns.heatmap(abs(self.data.corr()) < threshold, annot=False, cmap='Greys')
            fig.savefig(self.folder + 'CoLinearity/v%i/' % self.version + self.tag + 'Matrix.png',
                        format='png', dpi=300)

            # Minimum representation
            corr_mat = self.data.corr().abs()
            upper = corr_mat.where(np.triu(np.ones(corr_mat.shape), k=1).astype('bool'))
            col_drop = [column for column in upper.columns if any(upper[column] > 0.95)]
            minimal_rep = self.data.drop(self.data[col_drop], axis=1)
            fig = plt.figure(figsize=[24, 16])
            sns.heatmap(abs(minimal_rep.corr()) < threshold, annot=False, cmap='Greys')
            fig.savefig(self.folder + 'CoLinearity/v%i/' % self.version + self.tag + 'Minimum_Representation.png',
                        format='png', dpi=300)

    def differencing(self):
        if self.plotDifferencing:
            # Create folder
            if not os.path.exists(self.folder + 'Lags/'):
                os.mkdir(self.folder + 'Lags/')

            # Skip if existing
            if self.tag + 'Variance.png' in os.listdir(self.folder + 'Lags/'):
                return

            # Setup
            max_lags = 4
            var_vec = np.zeros((max_lags, len(self.data.keys())))
            diff_data = self.data / np.sqrt(self.data.var())

            # Calculate variance per lag
            for i in range(max_lags):
                var_vec[i, :] = diff_data.var()
                diff_data = diff_data.diff(1)[1:]

            # Plot
            fig = plt.figure(figsize=[24, 16])
            plt.title('Variance for different lags')
            plt.plot(var_vec)
            plt.xlabel('Lag')
            plt.yscale('log')
            plt.ylabel('Average variance')
            fig.savefig(self.folder + 'Lags/' + self.tag + 'Variance.png', format='png', dpi=300)

    def complete_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/ACF/'):
                os.makedirs(self.folder + 'Correlation/ACF/')

            # Difference data
            diff_data = copy.copy(self.data)
            for i in range(self.differ):
                diff_data = diff_data.diff(1)[1:]

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v%i.png' % self.version in \
                        os.listdir(self.folder + 'Correlation/ACF/'):
                    continue

                # Plot
                fig = plot_acf(diff_data[key], fft=True)
                plt.title(key)
                fig.savefig(self.folder + 'Correlation/ACF/' + self.tag + key + '_differ_' + str(self.differ) +
                            '_v%i.png' % self.version, format='png', dpi=300)
                plt.close()

    def partial_auto_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/PACF/'):
                os.makedirs(self.folder + 'Correlation/PACF/')

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v%i.png' % self.version in \
                        os.listdir(self.folder + 'EDA/Correlation/PACF/'):
                    continue

                # Plot
                try:
                    fig = plot_pacf(self.data[key])
                    fig.savefig(self.folder + 'EDA/Correlation/PACF/' + self.tag + key + '_differ_' +
                                str(self.differ) + '_v%i.png' % self.version, format='png', dpi=300)
                    plt.title(key)
                    plt.close()
                except Exception as e:
                    # todo find exception
                    raise ValueError(e)

    def cross_corr(self):
        if self.plotSignalCorrelations:
            # Create folder
            if not os.path.exists(self.folder + 'Correlation/Cross/'):
                os.makedirs(self.folder + 'Correlation/Cross/')

            # Prepare
            folder = 'Correlation/Cross/'
            y = self.Y.to_numpy().reshape((-1))

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if self.tag + key + '_differ_' + str(self.differ) + '_v%i.png' % self.version in \
                        os.listdir(self.folder + folder):
                    continue

                # Plot
                try:
                    fig = plt.figure(figsize=[24, 16])
                    plt.xcorr(self.data[key], y, maxlags=self.lags)
                    plt.title(key)
                    fig.savefig(self.folder + folder + self.tag + key + '_differ_' + str(self.differ) +
                                '_v%i.png' % self.version, format='png', dpi=300)
                    plt.close()
                except Exception as e:
                    raise ValueError(e)

    def scatters(self):
        if self.plotScatterPlots:
            # Create folder
            if not os.path.exists(self.folder + 'Scatters/v%i/' % self.version):
                os.makedirs(self.folder + 'Scatters/v%i/' % self.version)

            # Iterate through features
            for key in tqdm(self.data.keys()):
                # Skip if existing
                if '{}{}.png'.format(self.tag, key) in os.listdir(self.folder + 'Scatters/v%i/' % self.version):
                    continue

                # Plot
                fig = plt.figure(figsize=[24, 16])
                plt.scatter(self.Y, self.data[key], alpha=0.2)
                plt.ylabel(key)
                plt.xlabel('Output')
                plt.title('Scatter Plot ' + key + ' - Output')
                fig.savefig(self.folder + 'Scatters/v%i/' % self.version + self.tag + key + '.png',
                            format='png', dpi=100)
                plt.close(fig)

    def shap(self, args=None):
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v%i/' % self.version):
                os.makedirs(self.folder + 'Features/v%i/' % self.version)

            # Skip if existing
            if self.tag + 'SHAP.png' in os.listdir(self.folder + 'Features/v%i/' % self.version):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Calculate SHAP values
            import shap
            shap_values = shap.TreeExplainer(model).shap_values(self.data)

            # Plot
            fig = plt.figure(figsize=[8, 32])
            plt.subplots_adjust(left=0.4)
            shap.summary_plot(shap_values, self.data, plot_type='bar')
            fig.savefig(self.folder + 'Features/v%i/' % self.version + self.tag + 'SHAP.png', format='png', dpi=300)

    def feature_ranking(self, **args):
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v%i/' % self.version):
                os.mkdir(self.folder + 'Features/v%i/' % self.version)

            # Skip if existing
            if self.tag + 'RF.png' in os.listdir(self.folder + 'Features/v%i/' % self.version):
                return

            # Create model
            if self.mode == 'classification':
                model = RandomForestClassifier(**args).fit(self.data, self.Y)
            else:
                model = RandomForestRegressor(**args).fit(self.data, self.Y)

            # Plot
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            ind = np.argsort(model.feature_importances_)
            plt.barh(list(self.data.keys()[ind])[-15:], width=model.feature_importances_[ind][-15:],
                     color='#2369ec')
            fig.savefig(self.folder + 'Features/v%i/' % self.version + self.tag + 'RF.png', format='png', dpi=300)
            plt.close()

            # Store results
            results = pd.DataFrame({'x': self.data.keys(), 'score': model.feature_importances_})
            results.to_csv(self.folder + 'Features/v%i/' % self.version + self.tag + 'RF.csv')

    def predictive_power_score(self):
        if self.plotFeatureImportance:
            # Create folder
            if not os.path.exists(self.folder + 'Features/v%i/' % self.version):
                os.mkdir(self.folder + 'Features/v%i/' % self.version)

            # Skip if existing
            if self.tag + 'PPScore.png' in os.listdir(self.folder + 'Features/v%i/' % self.version):
                return

            # Calculate PPS
            data = self.data.copy()
            if isinstance(self.Y, pd.core.series.Series):
                data.loc[:, 'target'] = self.Y
            elif isinstance(self.Y, pd.DataFrame):
                data.loc[:, 'target'] = self.Y.loc[:, self.Y.keys()[0]]
            pp_score = ppscore.predictors(data, 'target').sort_values('ppscore')

            # Plot
            fig, ax = plt.subplots(figsize=[4, 6], constrained_layout=True)
            plt.subplots_adjust(left=0.5, top=1, bottom=0)
            ax.spines['right'].set_visible(False)
            ax.spines['bottom'].set_visible(False)
            ax.spines['top'].set_visible(False)
            plt.barh(pp_score['x'][-15:], width=pp_score['ppscore'][-15:], color='#2369ec')
            fig.savefig(self.folder + 'Features/v%i/' % self.version + self.tag + 'Ppscore.png', format='png', dpi=400)
            plt.close()

            # Store results
            pp_score.to_csv(self.folder + 'Features/v%i/pp_score.csv' % self.version)



import os
import json
import copy
import ppscore
import inspect
import functools
import numpy as np
import pandas as pd
from boruta import BorutaPy
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map
from sklearn.tree import DecisionTreeClassifier
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestClassifier
from sklearn.ensemble import RandomForestRegressor
from sklearn.cluster import MiniBatchKMeans


class FeatureProcessing:
    # todo implement time budget

    def __init__(self,
                 max_lags=10,
                 max_diff=2,
                 information_threshold=0.99,
                 extract_features=True,
                 folder='',
                 mode=None,
                 version=''):
        self.X = None
        self.originalInput = None
        self.Y = None
        self.model = None
        self.mode = mode
        self.threshold = None
        # Register
        self.baseScore = {}
        self.colinearFeatures = None
        self.crossFeatures = None
        self.kMeansFeatures = None
        self.diffFeatures = None
        self.laggedFeatures = None
        # Parameters
        self.maxLags = max_lags
        self.maxDiff = max_diff
        self.informationThreshold = information_threshold
        self.extractFeatures = extract_features
        self.folder = folder if folder == '' or folder[-1] == '/' else folder + '/'
        self.version = version
        # Tests
        assert 0 <= max_lags < 50, 'Max lags needs to be within [0, 50]'
        assert 0 <= max_diff < 3, 'Max diff needs to be within [0, 3]'
        assert 0 < information_threshold < 1, 'Information threshold needs to be within [0, 1]'
        assert mode is not None, 'Mode needs to be specified (regression or classification'

    def extract(self, input_frame, output_frame):
        self._clean_set(input_frame, output_frame)
        if self.extractFeatures:
            # Manipulate features
            # Order of these doesn't matter, all take originalInput features
            self._remove_co_linearity()
            self._calc_baseline()
            self._add_cross_features()
            self._add_k_means_features()
            self._add_trigonometry_features()
            self._add_diff_features()
            self._add_lagged_features()
        return self.X

    def select(self, input_frame, output_frame):
        # Check if not exists
        if os.path.exists(self.folder + 'Sets_v%i.json' % self.version):
            return json.load(open(self.folder + 'Sets_v%i.json' % self.version, 'r'))

        # Execute otherwise
        else:
            # Clean
            self._clean_set(input_frame, output_frame)

            # Different Feature Sets
            result = {'PPS': self._predictive_power_score()}
            result['RFT'], result['RFI'] = self._random_forest_importance()
            # result['BP'] = self._borutapy()

            # Store & Return
            json.dump(result, open(self.folder + 'Sets_v%i.json' % self.version, 'w'))
            return result

    def transform(self, data, features, **args):
        # Split Features
        cross_features = [k for k in features if '__x__' in k or '__d__' in k]
        linear_features = [k for k in features if '__sub__' in k or '__add__' in k]
        trigonometric_features = [k for k in features if 'sin__' in k or 'cos__' in k]
        k_means_features = [k for k in features if 'dist__' in k]
        diff_features = [k for k in features if '__diff__' in k]
        lag_features = [k for k in features if '__lag__' in k]
        original_features = [k for k in features if '__' not in k]

        # Make sure centers are provided if kMeansFeatures are nonzero
        if len(k_means_features) != 0:
            if 'k_means' not in args:
                raise ValueError('For K-Means features, the Centers need to be provided.')
            k_means = args['k_means']

        # Fill missing features for normalization
        required = copy.copy(original_features)
        required += [s.split('__')[::2] for s in cross_features]
        required += [s.split('__')[::2] for s in linear_features]
        required += [k.split('__')[1] for k in trigonometric_features]
        required += [s.split('__diff__')[0] for s in diff_features]
        required += [s.split('__lag__')[0] for s in lag_features]
        if len(k_means_features) != 0:
            required += [k for k in k_means.keys()]
        for k in [k for k in required if k not in data.keys()]:
            data.loc[:, k] = np.zeros(len(data))

        # Select
        X = data[original_features]

        # Multiplicative features
        for key in cross_features:
            if '__x__' in key:
                key_a, key_b = k.split('__x__')
                feature = X[key_a] * X[key_b]
                X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)
            else:
                key_a, key_b = k.split('__d__')
                feature = X[key_a] / X[key_b]
                X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)

        # Linear features
        for key in linear_features:
            if '__sub__' in key:
                key_a, key_b = key.split('__sub__')
                feature = X[key_a] - X[key_b]
                X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)
            else:
                key_a, key_b = key.split('__add__')
                feature = X[key_a] + X[key_b]
                X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)

        # Differentiated features
        for k in diff_features:
            key, diff = k.split('__diff__')
            feature = data[key]
            for i in range(1, diff):
                feature = feature.diff().fillna(0)
            X.loc[:, k] = feature

        # K-Means features
        if len(k_means_features) != 0:
            # Organise data
            temp = copy.copy(data)
            centers = k_means.iloc[:-2]
            means = k_means.iloc[-2]
            stds = k_means.iloc[-1]
            # Normalize
            temp -= means
            temp /= stds
            # Calculate centers
            for key in k_means_features:
                ind = int(key[key.find('dist__') + 6: key.rfind('_')])
                X.loc[:, key] = np.sqrt(np.square(temp.loc[:, centers.keys()] - centers.iloc[ind]).sum(axis=1))

        # Lagged features
        for k in lag_features:
            key, lag = k.split('__lag__')
            X.loc[:, key] = data[key].shift(-int(lag), fill_value=0)

        # Trigonometric features
        for k in trigonometric_features:
            func, key = k.split('__')
            self.X[k] = getattr(np, func)(self.X[key])

        return X

    def export_function(self):
        code = inspect.getsource(self.transform)
        return """

        ############
        # Features #
        ############""" + code[code.find('\n'): code.rfind('\n', 0, code.rfind('\n'))]

    def _clean_set(self, input_frame, output_frame):
        assert isinstance(input_frame, pd.DataFrame), 'Input supports only Pandas DataFrame'
        assert isinstance(output_frame, pd.Series), 'Output supports only Pandas Series'
        if self.mode == 'classification':
            self.model = DecisionTreeClassifier(max_depth=3)
            if output_frame.nunique() > 2:
                self.mode = 'multiclass'
        elif self.mode == 'regression':
            self.model = DecisionTreeRegressor(max_depth=3)
        # Bit of necessary data cleaning (shouldn't change anything)
        input_frame = input_frame.astype('float32').replace(np.inf, 1e12).replace(-np.inf, -1e12).fillna(0)\
            .reset_index(drop=True)
        self.X = copy.copy(input_frame)
        self.originalInput = copy.copy(input_frame)
        self.Y = output_frame.replace([np.inf, -np.inf], 0).fillna(0).reset_index(drop=True)

    def _calc_baseline(self):
        baseline = {}
        for key in self.originalInput.keys():
            baseline[key] = self._analyse_feature(self.X[key])
        self.baseline = pd.DataFrame(baseline).max(axis=1)

    def _analyse_feature(self, feature):
        feature = feature.clip(lower=1e-12, upper=1e12).fillna(0).values.reshape((-1, 1))
        m = copy.copy(self.model)
        m.fit(feature, self.Y)
        if self.mode == 'multiclass':
            return [m.score(feature, self.Y, self.Y == i) for i in self.Y.unique()]
        else:
            return [m.score(feature, self.Y)]

    def _accept_feature(self, candidate):
        if (candidate > self.baseline.values).any():
            return True
        else:
            return False

    @ staticmethod
    def _select_features(scores):
        # Convert dict to dataframe
        scores = pd.DataFrame(scores)

        # Select inds
        features_per_bin = 50
        inds = []
        for score in range(len(scores)):
            inds += [i for i, k in enumerate(scores.keys())
                     if k in scores.loc[score].sort_values(ascending=False).keys()[:features_per_bin]]

        # Return Keys
        return list(scores.keys()[np.unique(inds)])

    def _remove_co_linearity(self):
        """
        Calculates the Pearson Correlation Coefficient for all input features.
        Those higher than the information threshold are linearly codependent (i.e., describable by y = a x + b)
        These features add little to no information and are therefore removed.
        """
        # Check if not already executed
        if os.path.exists(self.folder + 'Colinear_v%i.json' % self.version):
            print('[Features] Loading Colinear features')
            self.colinearFeatures = json.load(open(self.folder + 'Colinear_v%i.json' % self.version, 'r'))

        # Else run
        else:
            print('[Features] Analysing colinearity')
            nk = len(self.X.keys())
            norm = (self.X - self.X.mean(skipna=True, numeric_only=True)).to_numpy()
            ss = np.sqrt(np.sum(norm ** 2, axis=0))
            corr_mat = np.zeros((nk, nk))
            for i in range(nk):
                for j in range(nk):
                    if i == j:
                        continue
                    if corr_mat[i, j] == 0:
                        c = abs(np.sum(norm[:, i] * norm[:, j]) / (ss[i] * ss[j]))
                        corr_mat[i, j] = c
                        corr_mat[j, i] = c
            upper = np.triu(corr_mat)
            self.colinearFeatures = self.X.keys()[np.sum(upper > self.informationThreshold, axis=0) > 0].to_list()
            json.dump(self.colinearFeatures, open(self.folder + 'Colinear_v%i.json' % self.version, 'w'))

        self.originalInput = self.originalInput.drop(self.colinearFeatures, axis=1)
        print('[Features] Removed %i Co-Linear features (%.3f %% threshold)' % (
            len(self.colinearFeatures), self.informationThreshold))

    @staticmethod
    def _static_multiply(model, data, labels, features):
        feature = data[features[0]] * data[features[1]]
        feature = feature.clip(lower=1e-12, upper=1e12).fillna(0).values.reshape((-1, 1))
        model.fit(feature, labels)
        return (features[0] + '__x__' + features[1],
                max([model.score(feature, labels, labels == i) for i in labels.unique()]))

    def _add_multi_features_mp(self):
        print('[Features] Listing Multiplication Features')
        features = []
        for i, key_a in enumerate(self.originalInput.keys()):
            for j, key_b in enumerate(self.originalInput.keys()):
                if i >= j:
                    continue
                features.append((key_a, key_b))
        print('[Features] Analysing {} Multiplication Features'.format(len(features)))
        scores = dict(process_map(functools.partial(self._static_multiply, self.model, self.X, self.Y),
                                  features, max_workers=8, chunksize=min(100, int(len(features) / 8 / 8))))
        self.multiFeatures = self._select_features(scores)

    def _add_cross_features(self):
        """
        Calculates cross-feature features with m and multiplication.
        Should be limited to say ~500.000 features (runs about 100-150 features / second)
        """
        # Check if not already executed
        if os.path.exists(self.folder + 'crossFeatures_v%i.json' % self.version) and False:
            self.crossFeatures = json.load(open(self.folder + 'crossFeatures_v%i.json' % self.version, 'r'))
            print('[Features] Loaded %i cross features' % len(self.crossFeatures))

        # Else, execute
        else:
            print('[Features] Analysing cross features')
            scores = {}
            n_keys = len(self.originalInput.keys())

            # Analyse Cross Features
            for i, key_a in enumerate(tqdm(self.originalInput.keys())):
                accepted_for_key_a = 0
                for j, key_b in enumerate(self.originalInput.keys()[np.random.permutation(n_keys)]):
                    # Skip if they're the same
                    if key_a == key_b:
                        continue
                    # Skip rest if key_a is not useful in first min(50, 30%) (uniform)
                    if accepted_for_key_a == 0 and j > min(50, int(n_keys / 3)):
                        continue

                    # Analyse Division
                    feature = self.X[key_a] / self.X[key_b]
                    score = self._analyse_feature(feature)
                    # Accept or not
                    if self._accept_feature(score):
                        scores[key_a + '__d__' + key_b] = score
                        accepted_for_key_a += 1

                    # Multiplication i * j == j * i, so skip if j >= i
                    if j >= i:
                        continue

                    # Analyse Multiplication
                    feature = self.X[key_a] * self.X[key_b]
                    score = self._analyse_feature(feature)
                    # Accept or not
                    if self._accept_feature(score):
                        scores[key_a + '__x__' + key_b] = score
                        accepted_for_key_a += 1

            # Select valuable features
            self.crossFeatures = self._select_features(scores)

        # Add features
        for k in self.crossFeatures:
            if '__x__' in k:
                key_a, key_b = k.split('__x__')
                feature = self.X[key_a] * self.X[key_b]
                self.X[k] = feature.clip(lower=1e-12, upper=1e12).fillna(0)
            else:
                key_a, key_b = k.split('__d__')
                feature = self.X[key_a] / self.X[key_b]
                self.X[k] = feature.clip(lower=1e-12, upper=1e12).fillna(0)

        # Store
        json.dump(self.crossFeatures, open(self.folder + 'crossFeatures_v%i.json' % self.version, 'w'))
        print('[Features] Added %i cross features' % len(self.crossFeatures))

    def _add_trigonometry_features(self):
        """
        Calculates trigonometry features with sinus, cosines
        """
        # Check if not already executed
        if os.path.exists(self.folder + 'trigoFeatures_v%i.json' % self.version):
            self.trigoFeatures = json.load(open(self.folder + 'trigoFeatures_v%i.json' % self.version, 'r'))
            print('[Features] Loaded %i trigonometric features' % len(self.trigoFeatures))

        # Else, execute
        else:
            print('[Features] Analysing Trigonometric features')

            scores = {}
            for key in tqdm(self.originalInput.keys()):
                # Sinus feature
                sin_feature = np.sin(self.X[key])
                score = self._analyse_feature(sin_feature)
                if self._accept_feature(score):
                    score['sin__' + key] = score

                # Cosinus feature
                cos_feature = np.cos(self.X[key])
                score = self._analyse_feature(cos_feature)
                if self._accept_feature(score):
                    scores['cos__' + key] = score

            # Select valuable features
            self.trigoFeatures = self._select_features(scores)

        # Add features
        for k in self.trigoFeatures:
            func, key = k.split('__')
            self.X[k] = getattr(np, func)(self.X[key])

        # Store
        json.dump(self.trigoFeatures, open(self.folder + 'trigoFeatures_v%i.json' % self.version, 'w'))
        print('[Features] Added %i trigonometric features' % len(self.trigoFeatures))

    def _add_additive_features(self):
        """
        Calculates simple additive and subtractive features
        """
        # Load if available
        if os.path.exists(self.folder + 'addFeatures_v%i.json' % self.version):
            self.addFeatures = json.load(open(self.folder + 'addFeatures_v%i.json' % self.version, 'r'))
            print('[Features] Loaded %i additive features' % len(self.addFeatures))

        # Else, execute
        else:
            scores = {}
            n_keys = len(self.originalInput.keys())
            for i, key_a in enumerate(self.originalInput.keys()):
                accepted_for_key_a = 0
                for j, key_b in enumerate(self.originalInput.keys()[np.random.permutation(n_keys)]):
                    # Skip if they're the same
                    if key_a == key_b:
                        continue
                    # Skip rest if key_a is not useful in first min(50, 30%) (uniform)
                    if accepted_for_key_a == 0 and j > min(50, int(n_keys / 3)):
                        continue

                    # Subtracting feature
                    feature = self.X[key_a] - self.X[key_b]
                    score = self._analyse_feature(feature)
                    if self._accept_feature(score):
                        scores[key_a + '__sub__' + key_b] = score
                        accepted_for_key_a += 1

                    # A + B == B + A, so skip if i > j
                    if j > i:
                        continue

                    # Additive feature
                    feature = self.X[key_a] + self.X[key_b]
                    score = self._analyse_feature(feature)
                    if self._accept_feature(score):
                        scores[key_a + '__add__' + key_b] = score
                        accepted_for_key_a += 1

            # Select valuable Features
            self.addFeatures = self._select_features(scores)

        # Add features
        for key in self.addFeatures:
            if '__sub__' in key:
                key_a, key_b = key.split('__sub__')
                feature = self.X[key_a] - self.X[key_b]
                self.X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)
            else:
                key_a, key_b = key.split('__add__')
                feature = self.X[key_a] + self.X[key_b]
                self.X.loc[:, key] = feature.clip(lower=1e-12, upper=1e12).fillna(0)

        # store
        json.dump(self.addFeatures, open(self.folder + 'addFeatures_v%i.json' % self.version, 'w'))
        print('[Features] Added %i additive features' % len(self.addFeatures))

    def _add_k_means_features(self):
        """
        Analyses the correlation of k-means features.
        k-means is a clustering algorithm which clusters the data.
        The distance to each cluster is then analysed.
        """
        # Check if not exist
        if os.path.exists(self.folder + 'K-MeansFeatures_v%i.json' % self.version):
            # Load features and cluster size
            self.kMeansFeatures = json.load(open(self.folder + 'K-MeansFeatures_v%i.json' % self.version, 'r'))
            k_means_data = pd.read_csv(self.folder + 'KMeans_v%i.csv' % self.version)
            print('[Features] Loaded %i K-Means features' % len(self.kMeansFeatures))

            # Prepare data
            data = copy.copy(self.originalInput)
            centers = k_means_data.iloc[:-2]
            means = k_means_data.iloc[-2]
            stds = k_means_data.iloc[-1]
            data -= means
            data /= stds

            # Add them
            for key in self.kMeansFeatures:
                ind = int(key[key.find('dist__') + 6: key.rfind('_')])
                self.X[key] = np.sqrt(np.square(data - centers.iloc[ind]).sum(axis=1))

        # If not executed, analyse all
        else:
            print('[Features] Calculating and Analysing K-Means features')
            # Prepare data
            data = copy.copy(self.originalInput)
            means = data.mean()
            stds = data.std()
            stds[stds == 0] = 1
            data -= means
            data /= stds

            # Determine clusters
            clusters = min(max(int(np.log10(len(self.originalInput)) * 8), 8), len(self.originalInput.keys()))
            kmeans = MiniBatchKMeans(n_clusters=clusters)
            column_names = ['dist__%i_%i' % (i, clusters) for i in range(clusters)]
            distances = pd.DataFrame(columns=column_names, data=kmeans.fit_transform(data))
            distances = distances.clip(lower=1e-12, upper=1e12).fillna(0)

            # Analyse correlation
            scores = {}
            for key in tqdm(distances.keys()):
                score = self._analyse_feature(distances[key])
                if self._accept_feature(score):
                    scores[key] = score

            # Add the valuable features
            self.kMeansFeatures = self._select_features(scores)
            for k in self.kMeansFeatures:
                self.X[k] = distances[k]

            # Create output
            centers = pd.DataFrame(columns=self.originalInput.keys(), data=kmeans.cluster_centers_)
            centers = centers.append(means, ignore_index=True)
            centers = centers.append(stds, ignore_index=True)
            centers.to_csv(self.folder + 'KMeans_v%i.csv' % self.version, index=False)
            json.dump(self.kMeansFeatures, open(self.folder + 'K-MeansFeatures_v%i.json' % self.version, 'w'))
            print('[Features] Added %i K-Means features (%i clusters)' % (len(self.kMeansFeatures), clusters))

    def _add_diff_features(self):
        """
        Analyses whether the differenced signal of the data should be included.
        """

        # Check if we're allowed
        if self.maxDiff == 0:
            print('[Features] Differenced features skipped, max diff = 0')
            return

        # Check if exist
        if os.path.exists(self.folder + 'diffFeatures_v%i.json' % self.version):
            self.diffFeatures = json.load(open(self.folder + 'diffFeatures_v%i.json' % self.version, 'r'))
            print('[Features] Loaded %i differenced features' % len(self.diffFeatures))

        # If not exist, execute
        else:
            print('[Features] Analysing differenced features')
            # Copy data so we can diff without altering original data
            keys = self.originalInput.keys()
            diff_input = copy.copy(self.originalInput)

            # Calculate scores
            scores = {}
            for diff in tqdm(range(1, self.maxDiff + 1)):
                diff_input = diff_input.diff().fillna(0)
                for key in keys:
                    score = self._analyse_feature(diff_input[key])
                    if self._accept_feature(score):
                        scores[key + '__diff__%i' % diff] = score

            # Select the valuable features
            self.diffFeatures = self._select_features(scores)
            print('[Features] Added %i differenced features' % len(self.diffFeatures))

        # Add Differenced Features
        for k in self.diffFeatures:
            key, diff = k.split('__diff__')
            feature = self.X[key]
            for i in range(1, diff):
                feature = feature.diff().clip(lower=1e-12, upper=1e12).fillna(0)
            self.X[k] = feature
        json.dump(self.diffFeatures, open(self.folder + 'diffFeatures_v%i.json' % self.version, 'w'))

    def _add_lagged_features(self):
        """
        Analyses the correlation of lagged features (value of sensor_x at t-1 to predict target at t)
        """
        # Check if allowed
        if self.maxLags == 0:
            print('[Features] Lagged features skipped, max lags = 0')
            return

        # Check if exists
        if os.path.exists(self.folder + 'laggedFeatures_v%i.json' % self.version):
            self.laggedFeatures = json.load(open(self.folder + 'laggedFeatures_v%i.json' % self.version, 'r'))
            print('[Features] Loaded %i lagged features' % len(self.laggedFeatures))

        # Else execute
        else:
            print('[Features] Analysing lagged features')
            keys = self.originalInput.keys()
            scores = {}
            for lag in tqdm(range(1, self.maxLags)):
                for key in keys:
                    score = self._analyse_feature(self.X[key].shift(-1))
                    if self._accept_feature(score):
                        scores[key + '__lag__%i' % lag] = score

            # Select
            self.laggedFeatures = self._select_features(scores)
            print('[Features] Added %i lagged features' % len(self.laggedFeatures))

        # Add selected
        for k in self.laggedFeatures:
            key, lag = k.split('__lag__')
            self.X[k] = self.originalInput[key].shift(-int(lag), fill_value=0)

        # Store
        json.dump(self.laggedFeatures, open(self.folder + 'laggedFeatures_v%i.json' % self.version, 'w'))

    def _predictive_power_score(self):
        """
        Calculates the Predictive Power Score (https://github.com/8080labs/ppscore)
        Assymmetric correlation based on single decision trees trained on 5.000 samples with 4-Fold validation.
        """
        print('[Features] Determining features with PPS')
        data = self.X.copy()
        data['target'] = self.Y.copy()
        pp_score = ppscore.predictors(data, "target")
        pp_cols = pp_score['x'][pp_score['ppscore'] != 0].to_list()
        print('[Features] Selected %i features with Predictive Power Score' % len(pp_cols))
        return pp_cols

    def _random_forest_importance(self):
        """
        Calculates Feature Importance with Random Forest, aka Mean Decrease in Gini Impurity
        Symmetric correlation based on multiple features and multiple trees ensemble
        """
        print('[Features] Determining features with RF')
        if self.mode == 'regression':
            rf = RandomForestRegressor().fit(self.X, self.Y)
        elif self.mode == 'classification' or self.mode == 'multiclass':
            rf = RandomForestClassifier().fit(self.X, self.Y)
        else:
            raise ValueError('Method not implemented')
        fi = rf.feature_importances_
        sfi = fi.sum()
        ind = np.flip(np.argsort(fi))
        # Info Threshold
        ind_keep = [ind[i] for i in range(len(ind)) if fi[ind[:i]].sum() <= self.informationThreshold * sfi]
        threshold = self.X.keys()[ind_keep].to_list()
        ind_keep = [ind[i] for i in range(len(ind)) if fi[i] > sfi / 100]
        increment = self.X.keys()[ind_keep].to_list()
        print('[Features] Selected %i features with RF thresholded' % len(threshold))
        print('[Features] Selected %i features with RF increment' % len(increment))
        return threshold, increment

    def _borutapy(self):
        print('[Features] Determining features with Boruta')
        if self.mode == 'regression':
            rf = RandomForestRegressor()
        elif self.mode == 'classification' or self.mode == 'multiclass':
            rf = RandomForestClassifier()
        selector = BorutaPy(rf, n_estimators='auto', verbose=0)
        selector.fit(self.X.to_numpy(), self.Y.to_numpy())
        bp_cols = self.X.keys()[selector.support_].to_list()
        print('[Features] Selected %i features with Boruta' % len(bp_cols))
        return bp_cols



import numpy as np
import pandas as pd


class Sequence:

    def __init__(self, back=0, forward=0, shift=0, diff='none'):
        """
        Sequencer. Sequences and differentiate data.
        Scenarios:
        - Sequenced I/O                     --> back & forward in int or list of ints
        - Sequenced input, single output    --> back: int, forward: list of ints
        - Single input, single output       --> back & forward =

        :param back: Int or List[int]: input indices to include
        :param forward: Int or List[int]: output indices to include
        :param shift: Int: If there is a shift between input and output
        :param diff: differencing algo, pick between 'none', 'diff', 'log_diff', 'fractional' (no revert)
        """
        if type(back) == int:
            back = np.linspace(0, back, back+1).astype('int')
            self.inputDType = 'int'
        elif type(back) == list:
            back = np.array(back)
            self.inputDType = 'list'
        else:
            raise ValueError('Back needs to be int or list(int)')
        if type(forward) == int:
            self.outputDType = 'int'
            forward = np.linspace(0, forward, forward+1).astype('int')
        elif type(forward) == list:
            self.outputDType = 'list'
            forward = np.array(forward)
        else:
            raise ValueError('Forward needs to be int or list(int)')
        self.backVec = back
        self.foreVec = forward
        self.nBack = len(back)
        self.nFore = len(forward)
        self.foreRoll = np.roll(self.foreVec, 1)
        self.foreRoll[0] = 0
        self.mBack = max(back) - 1
        self.mFore = max(forward) - 1
        self.shift = shift
        self.diff = diff
        self.samples = 1
        if diff != 'none':
            self.samples = 0
            self.nBack -= 1
        if diff not in ['none', 'diff', 'log_diff']:
            raise ValueError('Type should be in [None, diff, log_diff, fractional]')

    def convert(self, x, y):
        if isinstance(x, pd.DataFrame) or isinstance(x, pd.core.series.Series):
            assert isinstance(y, pd.DataFrame) or isinstance(y, pd.core.series.Series), \
                'Input and Output need to be the same data type.'
            return self.convert_pandas(x, y)
        elif isinstance(x, np.ndarray):
            assert isinstance(y, np.ndarray), 'Input and Output need to be the same data type.'
            return self.convert_numpy(x, y)
        else:
            TypeError('Input & Output need to be same datatype, either Numpy or Pandas.')

    def convert_numpy(self, x, y):
        # todo implement flat
        if x.ndim == 1:
            x = x.reshape((-1, 1))
        if y.ndim == 1:
            y = y.reshape((-1, 1))
        samples = len(x) - self.mBack - self.mFore - self.shift - 1
        features = len(x[0])
        input_sequence = np.zeros((samples, self.nBack, features))
        output_sequence = np.zeros((samples, self.nFore))
        if self.diff == 'none':
            for i in range(samples):
                input_sequence[i] = x[i + self.backVec]
                output_sequence[i] = y[i - 1 + self.mBack + self.shift + self.foreVec].reshape((-1))
            return input_sequence, output_sequence
        elif self.diff[-4:] == 'diff':
            if self.diff == 'log_diff':
                x = np.log(x)
                y = np.log(y)
            if (self.backVec == 0).all():
                self.backVec = np.array([0, 1])
            for i in range(samples):
                input_sequence[i] = x[i + self.backVec[1:]] - x[i + self.backVec[:-1]]
                output_sequence[i] = (y[i + self.mBack + self.shift + self.foreVec] -
                                      y[i + self.mBack + self.shift + self.foreRoll]).reshape((-1))
            return input_sequence, output_sequence

    def convert_pandas(self, x, y):
        # Check inputs
        if isinstance(x, pd.core.series.Series):
            x = x.to_frame()
        if isinstance(y, pd.core.series.Series):
            y = y.to_frame()
        assert len(x) == len(y)
        assert isinstance(x, pd.DataFrame)
        assert isinstance(y, pd.DataFrame)

        # Keys
        input_keys = x.keys()
        output_keys = y.keys()

        # No Differencing
        if self.diff == 'none':
            # Input
            for lag in self.backVec:
                keys = [key + '_' + str(lag) for key in input_keys]
                x[keys] = x[input_keys].shift(lag)

            # Output
            for shift in self.foreVec:
                keys = [key + '_' + str(shift) for key in output_keys]
                y[keys] = y[keys].shift(-shift)

        # With differencing
        elif self.diff[-4:] == 'diff':
            # Input
            for lag in self.backVec:
                # Shifted
                keys = [key + '_' + str(lag) for key in input_keys]
                x[keys] = x[input_keys].shift(lag)

                # differentiated
                d_keys = [key + '_d_' + str(lag) for key in input_keys]
                x[d_keys] = x[input_keys].shift(lag) - x[input_keys]

            # Output
            for shift in self.foreVec:
                # Only differentiated
                keys = [key + '_' + str(shift) for key in output_keys]
                y[keys] = y[output_keys].shift(shift) - y[output_keys]

        # Drop _0 (same as original)
        x = x.drop([key for key in x.keys() if '_0' in key], axis=1)
        y = y.drop([key for key in y.keys() if '_0' in key], axis=1)

        # Return (first lags are NaN, last shifts are NaN
        return x.iloc[lag:-shift if shift > 0 else None], y.iloc[lag:-shift if shift > 0 else None]

    def revert(self, differentiated, original):
        # Revert the sequencing loop: d = np.h stack((d[0], d[0] + np.cum sum(dd)))
        if self.nFore == 1:
            differentiated = differentiated.reshape((-1, 1))
        y = np.zeros_like(differentiated)
        if self.diff == 'none':
            return
        if self.diff == 'log_diff':
            for i in range(self.nFore):
                y[:, i] = np.log(original[self.mBack + self.foreRoll[i]:-self.foreVec[i]]) + differentiated[:, i]
            return np.exp(y)
        if self.diff == 'diff':
            for i in range(self.nFore):
                y[:, i] = original[self.mBack + self.foreRoll[i]:-self.foreVec[i]] + differentiated[:, i]
            return y



import re
import json


def boolean_input(question):
    x = input(question + ' [y / n]')
    if x.lower() == 'n' or x.lower() == 'no':
        return False
    elif x.lower() == 'y' or x.lower() == 'yes':
        return True
    else:
        print('Sorry, I did not understand. Please answer with "n" or "y"')
        return boolean_input(question)


def clean_keys(data):
    # Clean Keys
    new_keys = {}
    for key in data.keys():
        new_keys[key] = re.sub('[^a-zA-Z0-9 \n]', '_', key.lower()).replace('__', '_')
    data = data.rename(columns=new_keys)
    return data


def parse_json(json_string):
    if isinstance(json_string, dict):
        return json_string
    else:
        try:
            return json.loads(json_string
                              .replace("'", '"')
                              .replace("True", "true")
                              .replace("False", "false")
                              .replace("nan", "NaN")
                              .replace("None", "null"))
        except json.decoder.JSONDecodeError:
            print('[AutoML] Cannot validate, impassable JSON.')
            print(json_string)
            return json_string



import pandas as pd
import xgboost as xgb
from sklearn.model_selection import train_test_split


class XGBClassifier:

    def __init__(self, params=None):
        """
        XG Boost wrapper
        @param params: Model parameters
        """
        # Args
        self.params = params

        # Parameters
        self.model = None
        self.trained = False

    def fit(self, x, y, callback=None):
        # Split & Convert data
        train_x, test_x, train_y, test_y = train_test_split(x, y, test_size=0.1)
        d_train = xgb.DMatrix(train_x, label=train_y)
        d_test = xgb.DMatrix(test_x, label=test_y)

        # Model training
        self.model = xgb.train(self.params,
                               d_train,
                               evals=[(d_test, 'validation')],
                               callbacks=[callback])
        self.trained = True

    def predict(self, x):
        assert self.trained is True, 'Model not yet trained'
        # Convert if dataframe
        if isinstance(x, pd.DataFrame):
            x = x.to_numpy()
        # Convert if single column
        if len(x.shape) == 1:
            x = x.reshape((-1, 1))
        # Convert to DMatrix
        d_predict = xgb.DMatrix(x)
        self.model.predict(d_predict)

    def predict_proba(self, x):
        assert self.trained is True, 'Model not yet trained'
        # Convert if dataframe
        if isinstance(x, pd.DataFrame):
            x = x.to_numpy()
        # Convert if single column
        if len(x.shape) == 1:
            x = x.reshape((-1, 1))
        # Convert to DMatrix
        d_predict = xgb.DMatrix(x)
        self.model.predict_proba(d_predict)

    def set_params(self, params):
        self.params = params

    def get_params(self):
        return self.params



import pandas as pd
from sklearn.experimental import enable_halving_search_cv
from sklearn.model_selection import HalvingRandomSearchCV
import multiprocessing as mp
from .BaseGridSearch import BaseGridSearch


class HalvingGridSearch(BaseGridSearch):

    def __init__(self, model, params=None, cv=None, scoring=None, verbose=None):
        super().__init__(model, params, cv, scoring, verbose)
        # Params
        self.resource = 'n_samples'
        self.max_resource = 'auto'
        self.min_resource = 200

        # Update them
        self.set_resources()

    def set_resources(self):
        if self.model.__module__ == 'catboost.core':
            self.resource = 'n_estimators'
            self.max_resource = 3000
            self.min_resource = 250
        if self.model.__module__ == 'sklearn.ensemble._bagging' or \
                self.model.__module__ == 'xgboost.sklearn' or \
                self.model.__module__ == 'lightgbm.sklearn' or \
                self.model.__module__ == 'sklearn.ensemble._forest':
            self.resource = 'n_estimators'
            self.max_resource = 1500
            self.min_resource = 50

    def fit(self, x, y):
        # Update minimum resource for samples (based on dataset)
        if self.resource == 'n_samples':
            self.min_resource = int(0.2 * len(x)) if len(x) > 5000 else len(x)

        # Set up and run grid search
        halving_random_search = HalvingRandomSearchCV(self.model, self.params,
                                                      n_candidates=200,
                                                      resource=self.resource,
                                                      max_resources=self.max_resource,
                                                      min_resources=self.min_resource,
                                                      cv=self.cv,
                                                      scoring=self.scoring,
                                                      factor=3, n_jobs=mp.cpu_count() - 1,
                                                      verbose=self.verbose)
        halving_random_search.fit(x, y)

        # Parse results
        scikit_results = pd.DataFrame(halving_random_search.cv_results_)
        results = pd.DataFrame()
        results[['params', 'mean_objective', 'std_objective', 'mean_time', 'std_time']] = scikit_results[
            ['params', 'mean_test_score', 'std_test_score', 'mean_fit_time', 'std_fit_time']]
        results['worst_case'] = - results['mean_objective'] - results['std_objective']

        # Update resource in results
        if self.resource != 'n_samples':
            for i in range(len(results)):
                results.loc[results.index[i], 'params'][self.resource] = self.max_resource

        return results



import setuptools
import subprocess


version = (
    subprocess.run(["git", "describe", "--tags"], stdout=subprocess.PIPE)
    .stdout.decode("utf-8")
    .strip()
)
assert "." in version

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Amplo",
    version=version,
    author="Amplo GmbH",
    author_email="info@amplo.ch",
    description="Fully automated end to end machine learning pipeline",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nielsuit227/AutoML",
    packages=setuptools.find_packages(),
    package_data={"Amplo": ["VERSION"]},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
)



import time
import copy
import datetime
import itertools
import numpy as np
import pandas as pd
from tqdm import tqdm
from sklearn import metrics


class BaseGridSearch:

    def __init__(self, model, params=None, cv=None, scoring=metrics.SCORERS['accuracy'],
                 verbose=0):
        """
        Basic exhaustive grid search.
        @param model: Model object to optimize
        @param cv: Cross-Validator for scoring
        @param scoring: make_scorer for scoring
        """
        # Args
        self.model = model
        self.params = params
        self.cv = cv
        self.scoring = scoring
        self.verbose = verbose

        # Checks
        assert model is not None, 'Model not provided'
        assert isinstance(params, dict), 'Parameter set needs to be a dictionary'
        assert all([isinstance(x, list) for x in params.values()]), 'Parameter set dictionary needs to be filled with' \
                                                                    ' lists.'
        assert scoring in metrics.SCORERS.values(), 'Chose a scorer from sklearn.metrics.SCORERS'
        
        # Initiate
        self.parsedParams = []
        self.result = []
        self._parse_params()
        self.best = [None, None]    # Mean score, Std score

    def _parse_params(self):
        k, v = zip(*self.params.items())
        self.parsedParams = [dict(zip(k, v)) for v in itertools.product(*self.params.values())]
        print('[GridSearch] %i folds with %i parameter combinations, %i runs.' % (
            self.cv.n_splits,
            len(self.parsedParams),
            len(self.parsedParams) * self.cv.n_splits))

    def fit(self, x, y):
        # Convert to Numpy
        if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
            x = np.array(x)
        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y = np.array(y).reshape((-1))

        # Loop through parameters
        for i, param in tqdm(enumerate(self.parsedParams)):
            # print('[GridSearch] ', param)
            scoring = []
            timing = []
            for train_ind, val_ind in self.cv.split(x, y):
                # Start Timer
                t = time.time()

                # Split data
                x_train, x_val = x[train_ind], x[val_ind]
                y_train, y_val = y[train_ind], y[val_ind]

                # Model training
                model = copy.copy(self.model)
                model.set_params(**param)
                model.fit(x_train, y_train)

                # Results
                scoring.append(self.scoring(model, x_val, y_val))
                timing.append(time.time() - t)

            # Compare scores
            if np.mean(scoring) - np.std(scoring) > self.best[0] - self.best[1]:
                self.best = [np.mean(scoring), np.std(scoring)]

            if self.verbose > 0:
                print('[GridSearch][%s] Score: %.4f \u00B1 %.4f (in %.1f seconds) (Best score so-far: %.4f \u00B1 '
                      '%.4f) (%i / %i)' % (datetime.now().strftime('%H:%M'), np.mean(scoring), np.std(scoring),
                                           np.mean(timing), self.best[0], self.best[1], i + 1, len(self.parsedParams)))
            self.result.append({
                'scoring': scoring,
                'mean_objective': np.mean(scoring),
                'std_objective': np.std(scoring),
                'time': timing,
                'mean_time': np.mean(timing),
                'std_time': np.std(timing),
                'params': param
            })
        return pd.DataFrame(self.result)
